# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional, Sequence

from google.ads.googleads.client import GoogleAdsClient  # type: ignore
from importlib import import_module
from pathlib import Path
from proto.primitives import ProtoType


class BaseClient:

    def get_response(self, entity_id: str, query_text: str):
        pass


class GoogleAdsApiClient(BaseClient):
    default_google_ads_yaml = str(Path.home() / "google-ads.yaml")

    def __init__(self,
                 path_to_config: str = default_google_ads_yaml,
                 config_dict: Dict[str, Any] = None,
                 yaml_str: str = None,
                 version: str = "v9"):
        self.client = self._init_client(path=path_to_config,
                                        config_dict=config_dict,
                                        yaml_str=yaml_str,
                                        version=version)
        self.ads_service = self.client.get_service("GoogleAdsService")
        self.api_version = version

    def get_response(self, entity_id, query_text):
        response = self.ads_service.search_stream(customer_id=entity_id,
                                                  query=query_text)
        return response

    def _init_client(self, path, config_dict, yaml_str,
                     version) -> Optional[GoogleAdsClient]:
        if path:
            return GoogleAdsClient.load_from_storage(path, version)
        if config_dict:
            return GoogleAdsClient.load_from_dict(config_dict, version)
        if yaml_str:
            return GoogleAdsClient.load_from_string(yaml_str, version)
        try:
            return GoogleAdsClient.load_from_env(version)
        except Exception as e:
            raise ValueError("Cannot instantiate GoogleAdsClient")

    def infer_types(self, fields: Sequence[str]) -> Sequence[Any]:
        """Maps API fields to Python primitives."""

        base_module = f"google.ads.googleads.{self.api_version}"
        common_types_module = f"{base_module}.common.types"
        segments = import_module(f"{common_types_module}.segments")
        metrics = import_module(f"{common_types_module}.metrics")

        mapping = {"INT64": int, "FLOAT": float, "DOUBLE": float, "BOOL": bool}
        output = []
        for field in fields:
            resource, *sub_resource, base_field = field.split(".")
            if resource == "metrics":
                result = metrics.Metrics.meta.fields.get(
                    base_field).descriptor.type
            elif resource == "segments":
                # If segment has name segments.something.something
                if sub_resource:
                    target_segment = getattr(
                        segments, f"{clean_resource(sub_resource[-1])}")
                else:
                    target_segment = getattr(segments,
                                             f"{clean_resource(resource)}")
                result = target_segment.meta.fields.get(
                    base_field).descriptor.type
            else:
                resource_module = import_module(
                    f"{base_module}.resources.types.{resource}")

                target_resource = getattr(resource_module,
                                          f"{clean_resource(resource)}")
                try:
                    # If resource has name resource.something.something
                    if sub_resource:
                        target_resource = getattr(
                            target_resource,
                            f"{clean_resource(sub_resource[-1])}")
                except AttributeError:
                    resource_module = import_module(
                        f"{base_module}.resources.types.{sub_resource[0]}")
                    if len(sub_resource) > 1:
                        if hasattr(resource_module,
                                   f"{clean_resource(sub_resource[1])}"):
                            target_resource = getattr(
                                resource_module,
                                f"{clean_resource(sub_resource[-1])}")
                        else:
                            resource_module = import_module(
                                f"{common_types_module}.ad_type_infos")

                            target_resource = getattr(
                                resource_module,
                                f"{clean_resource(sub_resource[1])}Info")
                    else:
                        target_resource = getattr(
                            resource_module,
                            f"{clean_resource(sub_resource[-1])}")
                result = target_resource.meta.fields.get(
                    base_field).descriptor.type

            output.append(mapping.get(ProtoType(result).name, str))
        return output


def clean_resource(resource: str) -> str:
    return resource.title().replace('_', '')



from gaarf.bq_executor import extract_datasets


def test_extract_datasets():
    macros = {
        "start_date": ":YYYYMMDD",
        "bq_dataset": "dataset_1",
        "dataset_new": "dataset_2",
        "legacy_dataset_old": "dataset_3",
        "wrong_dts": "dataset_4"
    }

    expected = ["dataset_1", "dataset_2", "dataset_3"]
    datasets = extract_datasets(macros)
    assert datasets == expected



import pytest
import pandas as pd
from gaarf.report import GaarfReport, GaarfRow, GaarfIterator


@pytest.fixture
def single_element_report():
    return GaarfReport(results=[1], column_names=["campaign_id"])


@pytest.fixture
def single_column_report():
    return GaarfReport(results=[[1], [2], [3]], column_names=["campaign_id"])


@pytest.fixture
def multi_column_report():
    return GaarfReport(results=[[1, 2], [2, 3], [3, 4]],
                       column_names=["campaign_id", "ad_group_id"])


def test_single_element_report_returns_sequence(single_element_report):
    assert [row for row in single_element_report] == [1]


def test_single_column_report_returns_sequence(single_column_report):
    assert [row for row in single_column_report] == [1, 2, 3]


def test_multi_column_report_returns_gaarf_row(multi_column_report):
    results = [row for row in multi_column_report]
    assert isinstance(results[0], GaarfRow)


def test_multi_column_report_support_iteration_with_gaarf_iterator(
        multi_column_report):
    assert isinstance(iter(multi_column_report), GaarfIterator)


def test_multi_column_report_get_element_by_id(multi_column_report):
    assert [row[0] for row in multi_column_report] == [1, 2, 3]


def test_multi_column_report_get_element_by_name(multi_column_report):
    assert [row["campaign_id"] for row in multi_column_report] == [1, 2, 3]


def test_multi_column_report_get_element_with_get_method(multi_column_report):
    assert [row.get("campaign_id") for row in multi_column_report] == [1, 2, 3]


def test_multi_column_report_get_element_as_attribute(multi_column_report):
    assert [row.campaign_id for row in multi_column_report] == [1, 2, 3]


def test_getitem_raise_index_error(multi_column_report):
    with pytest.raises(IndexError):
        [row[99] for row in multi_column_report] == [1, 2, 3]


def test_get_return_non_value(multi_column_report):
    assert [row.get("missing_value")
            for row in multi_column_report] == [None, None, None]


def test_convert_report_to_pandas(multi_column_report):
    expected = pd.DataFrame(data=[[1, 2], [2, 3], [3, 4]],
                            columns=["campaign_id", "ad_group_id"])
    assert multi_column_report.to_pandas().equals(expected)


def test_get_report_length(multi_column_report):
    assert len(multi_column_report) == 3


def test_report_str(single_element_report):
    assert str(single_element_report) == "[1]"



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Sequence, Union
import pandas as pd


class GaarfReport:

    def __init__(self,
                 results: Sequence[Any],
                 column_names: Sequence[str],
                 is_fake: bool = False):
        self.results = results
        self.column_names = column_names
        self.is_fake = is_fake

    def to_list(self) -> Sequence[Any]:
        return self.results

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame(data=self.results, columns=self.column_names)

    def __len__(self):
        return len(self.results)

    def __iter__(self):
        return GaarfIterator(self.results, self.column_names)

    def __str__(self):
        return f"{self.results}"


class GaarfIterator:

    def __init__(self, results, column_names):

        self.results = results
        self.column_names = column_names
        self.single_column_report = len(self.column_names) == 1
        self.index = 0

    def __next__(self):
        try:
            result = self.results[self.index]
            if not isinstance(result, Sequence):
                result = [result]
        except IndexError as e:
            raise StopIteration from e
        self.index += 1
        if self.single_column_report:
            return result[0]
        return GaarfRow(result, self.column_names)


class GaarfRow:

    def __init__(self, data: Sequence[Union[int, float, str]],
                 column_names: Sequence[str]):
        self.data = data
        self.n_elements = len(data)
        self.column_names = column_names

    def __getattr__(self, element: str) -> Any:
        return self.data[self.column_names.index(element)]

    def __getitem__(self, element: Union[str, int]) -> Any:
        if isinstance(element, int) and element < self.n_elements:
            return self.data[element]
        if isinstance(element, str):
            return self.data[self.column_names.index(element)]
        raise IndexError(f"cannot find {element} element!")

    def get(self, item: str) -> Any:
        if item in self.column_names:
            return self.data[self.column_names.index(item)]
        return None



import pytest
from typing import Any, Dict, List
from dataclasses import dataclass
from proto import Message
import gaarf.parsers as parsers


@dataclass
class FakeGoogleAdsRowElement:
    value: int
    text: str
    name: str


@dataclass
class TextAttribute:
    text: str


@dataclass
class NameAttribute:
    name: str


@dataclass
class ValueAttribute:
    value: str


@dataclass
class FakeAdsRowMultipleElements:
    campaign_type: NameAttribute
    clicks: int
    resource: str
    value: str


@dataclass
class FakeQuerySpecification:
    customizers: Dict[str, Any]
    fields: List[str]


@pytest.fixture
def fake_google_ads_row_element():
    return FakeGoogleAdsRowElement(1, "2", "3")


@pytest.fixture
def google_ads_row_parser():
    return parsers.GoogleAdsRowParser()


def test_google_ads_row_parser_return_last_parser_in_chain(
        google_ads_row_parser):
    assert type(google_ads_row_parser.parser) == parsers.RepeatedParser


def test_google_ads_row_parser_chooses_correct_element_parser(
        google_ads_row_parser, fake_google_ads_row_element):
    assert google_ads_row_parser.parse(fake_google_ads_row_element) == "3"


def test_resource_formatter_get_resource():
    resource = parsers.ResourceFormatter.get_resource("name: id")
    assert resource == "id"


def test_resource_formatter_get_resource_id():
    resource = parsers.ResourceFormatter.get_resource_id(
        "customers/1/resource/2")
    assert resource == "2"


def test_resource_formatter_clear_resource_id_int():
    resource = parsers.ResourceFormatter.clean_resource_id('"1"')
    assert resource == 1


def test_resource_formatter_clear_resource_id_str():
    resource = parsers.ResourceFormatter.clean_resource_id('"value"')
    assert resource == "value"


@pytest.fixture
def base_parser():
    return parsers.BaseParser(None)


def test_base_parser(base_parser):
    assert base_parser.parse("") == None


@pytest.fixture
def text_attribute():
    return TextAttribute("some-text")


@pytest.fixture
def name_attribute():
    return NameAttribute("some-name")


@pytest.fixture
def value_attribute():
    return ValueAttribute(1)


def test_attribute_parser(base_parser, text_attribute, name_attribute,
                          value_attribute):
    attribute_parser = parsers.AttributeParser(base_parser)
    assert attribute_parser.parse("") == None
    assert attribute_parser.parse(name_attribute) == "some-name"
    assert attribute_parser.parse(text_attribute) == "some-text"
    assert attribute_parser.parse(value_attribute) == 1


@pytest.mark.skip("WIP")
def test_empty_attribute_parser(base_parser):
    empty_attribute_parser = parsers.EmptyAttributeParser(base_parser)
    assert empty_attribute_parser.parse("") == None
    message = Message()
    assert empty_attribute_parser.parse(message) == "Not set"


def test_google_ads_row_parser(google_ads_row_parser, text_attribute,
                               name_attribute, value_attribute):
    assert isinstance(google_ads_row_parser.parser, parsers.RepeatedParser)
    assert google_ads_row_parser.parse(name_attribute) == "some-name"
    assert google_ads_row_parser.parse(text_attribute) == "some-text"
    assert google_ads_row_parser.parse(value_attribute) == 1
    assert google_ads_row_parser.parse("some-value") == None


@pytest.fixture
def fake_ads_row():
    return FakeAdsRowMultipleElements(NameAttribute("SEARCH"), 1,
                                      "customers/1/resource/2",
                                      ValueAttribute(1))


@pytest.fixture
def fake_query_specification():
    customizers = {
        2: {
            "type": "resource_index",
            "value": 0
        },
        3: {
            "type": "nested_field",
            "value": "value"
        }
    }
    return FakeQuerySpecification(
        customizers, ["campaign_type", "clicks", "resource", "value"])


def test_get_attributes_from_row(google_ads_row_parser, fake_ads_row,
                                 fake_query_specification):
    extracted_rows = google_ads_row_parser._get_attributes_from_row(
        fake_ads_row, ["campaign_type", "clicks", "resource", "value"])
    assert extracted_rows == (NameAttribute("SEARCH"), 1,
                              "customers/1/resource/2", ValueAttribute(1))
    assert google_ads_row_parser.parse_ads_row(
        fake_ads_row, fake_query_specification) == ["SEARCH", 1, "2", 1]



from gaarf.api_clients import GoogleAdsApiClient


def test_infer_types():
    client = GoogleAdsApiClient(config_dict={}, version="v11")
    fields = [
        'ad_group_ad.ad.id',
        'ad_group_ad.ad.legacy_responsive_display_ad.long_headline',
        'customer.descriptive_name', 'campaign.id', 'campaign.name',
        'campaign.app_campaign_setting.bidding_strategy_goal_type',
        'segments.date', 'metrics.clicks', 'metrics.ctr',
        'segments.asset_interaction_target.interaction_on_this_asset',
    ]
    output = client.infer_types(fields)
    expected = [int, str, str, int, str, str, str, int, float, bool]
    assert output == expected



import pytest

from datetime import datetime
from dateutil.relativedelta import relativedelta
import gaarf.cli.utils as utils


def test_convert_date():
    current_date = datetime.today()
    current_year = datetime(current_date.year, 1, 1)
    current_month = datetime(current_date.year, current_date.month, 1)
    last_year = current_year - relativedelta(years=1)
    last_month = current_month - relativedelta(months=1)
    yesterday = current_date - relativedelta(days=1)

    non_macro_date = "2022-01-01"
    date_year = ":YYYY"
    date_month = ":YYYYMM"
    date_day = ":YYYYMMDD"
    date_year_minus_one = ":YYYY-1"
    date_month_minus_one = ":YYYYMM-1"
    date_day_minus_one = ":YYYYMMDD-1"

    non_macro_date_converted = utils.convert_date(non_macro_date)
    new_date_year = utils.convert_date(date_year)
    new_date_month = utils.convert_date(date_month)
    new_date_day = utils.convert_date(date_day)
    new_date_year_minus_one = utils.convert_date(date_year_minus_one)
    new_date_month_minus_one = utils.convert_date(date_month_minus_one)
    new_date_day_minus_one = utils.convert_date(date_day_minus_one)

    assert non_macro_date_converted == non_macro_date
    assert new_date_year == current_year.strftime("%Y-%m-%d")
    assert new_date_month == current_month.strftime("%Y-%m-%d")
    assert new_date_day == current_date.strftime("%Y-%m-%d")
    assert new_date_year_minus_one == last_year.strftime("%Y-%m-%d")
    assert new_date_month_minus_one == last_month.strftime("%Y-%m-%d")
    assert new_date_day_minus_one == yesterday.strftime("%Y-%m-%d")


def test_wrong_convert_date():
    date_day = ":YYYYMMDD-N"
    with pytest.raises(ValueError):
        utils.convert_date(date_day)


@pytest.fixture
def param_parser():
    return utils.ParamsParser(["macro", "sql", "template"])


def test_identify_param_pair_existing(param_parser):
    param_pair = param_parser._identify_param_pair(
        "macro", ["--macro.start_date", "2022-01-01"])
    assert param_pair == {"start_date": "2022-01-01"}


def test_identify_param_pair_empty(param_parser):
    param_pair = param_parser._identify_param_pair(
        "macro", ["--missing_param.start_date", "2022-01-01"])
    assert param_pair is None


def test_identify_param_pair_raises_error(param_parser):
    with pytest.raises(ValueError):
        param_parser._identify_param_pair(
            "macro", ["--macro.start_date", ":YYYYMMDD", "extra_element"])


@pytest.fixture
def current_date_iso():
    return datetime.today().strftime("%Y%m%d")


def test_parse_params(param_parser, current_date_iso):
    parsed_params = param_parser._parse_params(
        "macro",
        ["--macro.start_date=2022-01-01", "--macro.end_date=2022-12-31"])
    assert parsed_params == {
        "start_date": "2022-01-01",
        "end_date": "2022-12-31"
    }


def test_parse(param_parser, current_date_iso):
    parsed_params = param_parser.parse(
        ["--macro.start_date=2022-01-01", "--macro.end_date=2022-12-31"])
    assert parsed_params == {
        "macro": {
            "start_date": "2022-01-01",
            "end_date": "2022-12-31"
        },
        "sql": {},
        "template": {},
    }


@pytest.fixture
def config_args():
    from dataclasses import dataclass

    @dataclass
    class FakeConfig:
        customer_id: str
        save: str
        api_version: str
        project: str

    return FakeConfig("1", "console", "10", "fake-project")


@pytest.fixture
def config_saver(config_args):
    return utils.ConfigSaver("/tmp/config.yaml")


def test_gaarf_config_saver_gaarf_dont_save_empty_values(config_saver):
    gaarf_config = utils.GaarfConfig(output="console",
                                     api_version="10",
                                     account="1",
                                     params={},
                                     writer_params={},
                                     customer_ids_query="",
                                     customer_ids_query_file="")

    config = config_saver.prepare_config({}, gaarf_config)
    assert config == {
        "gaarf": {
            "account": "1",
            "output": "console",
            "api_version": "10",
        }
    }


def test_gaarf_config_saver_gaarf_dont_save_inner_empty_values(config_saver):
    gaarf_config = utils.GaarfConfig(output="console",
                                     api_version="10",
                                     account="1",
                                     params={"macro": {
                                         "start_date": ":YYYYMMDD"
                                     }},
                                     writer_params={},
                                     customer_ids_query="",
                                     customer_ids_query_file="")

    config = config_saver.prepare_config({}, gaarf_config)
    assert config == {
        "gaarf": {
            "account": "1",
            "output": "console",
            "api_version": "10",
            "params": {
                "macro": {
                    "start_date": ":YYYYMMDD"
                }
            }
        }
    }


def test_config_saver_gaarf_save_customer_ids_query_values(config_saver):
    gaarf_config = utils.GaarfConfig(output="console",
                                     api_version="10",
                                     account="1",
                                     params={},
                                     writer_params={},
                                     customer_ids_query="SELECT",
                                     customer_ids_query_file="path/to/file.sql")

    config = config_saver.prepare_config({}, gaarf_config)
    assert config == {
        "gaarf": {
            "account": "1",
            "output": "console",
            "api_version": "10",
            "customer_ids_query": "SELECT",
            "customer_ids_query_file": "path/to/file.sql"
        }
    }


def test_config_saver_gaarf_bq(config_saver):
    gaarf_bq_config = utils.GaarfBqConfig(
        project="fake-project",
        params={"macro": {"bq_project": "another-fake-project"}})
    config = config_saver.prepare_config({}, gaarf_bq_config)
    assert config == {
        "gaarf-bq": {
            "project": "fake-project",
            "params": {
                "macro": {
                    "bq_project": "another-fake-project"
                }
            }
        }
    }


def test_config_saver_does_not_save_empty_params(config_saver):
    gaarf_bq_config = utils.GaarfBqConfig(
        project="fake-project",
        params={})
    config = config_saver.prepare_config({}, gaarf_bq_config)
    assert config == {
        "gaarf-bq": {
            "project": "fake-project",
        }
    }


def test_config_saver_does_not_save_empty_nested_params(config_saver):
    gaarf_bq_config = utils.GaarfBqConfig(
        project="fake-project",
        params={"macro": {"bq_project": "another-fake-project"}, "sql": {}})
    config = config_saver.prepare_config({}, gaarf_bq_config)
    assert config == {
        "gaarf-bq": {
            "project": "fake-project",
            "params": {
                "macro": {
                    "bq_project": "another-fake-project"
                }
            }
        }
    }


@pytest.fixture
def config_with_runtime_params():
    return utils.GaarfConfig(output="console",
                             api_version="10",
                             account="1",
                             params={"macro": {
                                 "start_date": ":YYYYMMDD"
                             }},
                             writer_params={},
                             customer_ids_query=None,
                             customer_ids_query_file=None)


@pytest.fixture
def config_without_runtime_params():
    return utils.GaarfConfig(output="console",
                             api_version="10",
                             account="1",
                             params={"macro": {
                                 "start_date": "2022-01-01"
                             }},
                             writer_params={},
                             customer_ids_query=None,
                             customer_ids_query_file=None)


def test_initialize_config_with_runtime_parameters(config_with_runtime_params,
                                                   current_date_iso):
    initialized_config = utils.initialize_runtime_parameters(
        config_with_runtime_params)
    expected_config = utils.GaarfConfig(
        output="console",
        api_version="10",
        account="1",
        params={
            "macro": {
                "start_date": datetime.today().strftime("%Y-%m-%d"),
                "date_iso": current_date_iso
            }
        },
        writer_params={},
        customer_ids_query=None,
        customer_ids_query_file=None)
    assert initialized_config == expected_config


def test_initialize_config_without_runtime_parameters(
        config_without_runtime_params, current_date_iso):
    initialized_config = utils.initialize_runtime_parameters(
        config_without_runtime_params)
    expected_config = utils.GaarfConfig(output="console",
                                        api_version="10",
                                        account="1",
                                        params={
                                            "macro": {
                                                "start_date": "2022-01-01",
                                                "date_iso": current_date_iso
                                            }
                                        },
                                        writer_params={},
                                        customer_ids_query=None,
                                        customer_ids_query_file=None)
    assert initialized_config == expected_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Tuple
import dataclasses
import re


@dataclasses.dataclass
class QueryElements:
    """Contains raw query and parsed elements.

    Attributes:
        query_title: Title of the query that needs to be parsed.
        query_text: Text of the query that needs to be parsed.
        fields: Ads API fields that need to be feched.
        column_names: friendly names for fields which are used when saving data
        customizers: Attributes of fields that need to be be extracted.
    """
    query_title: str
    query_text: str
    fields: List[str]
    column_names: List[str]
    customizers: Optional[Dict[int, Dict[str, str]]]
    resource_name: str
    is_constant_resource: bool


class QuerySpecification:
    def __init__(self, text: str, title: str = None,
                 args: Dict[Any, Any] = None):
        self.text = text
        self.title = title
        self.args = args

    def generate(self) -> QueryElements:
        """Reads query from a file and returns different elements of a query.

        Args:
            path: Path to a file with a query.

        Returns:
            Various elements parsed from a query (text, fields,
            column_names, etc).
        """

        query_lines = self.cleanup_query_text(self.text)
        resource_name = self.extract_resource_from_query(self.text)
        is_constant_resource = bool(resource_name.endswith("_constant"))
        query_text = self.normalize_query(" ".join(query_lines))
        if self.args:
            if (macros := self.args.get("macro")):
                query_text = query_text.format(**macros)
        fields = []
        column_names = []
        customizers = {}

        field_index = 0
        query_lines = self.extract_query_lines(" ".join(query_lines))
        for line in query_lines:
            field_elements, alias = self.extract_fields_and_aliases(line)
            field_name, customizer_type, customizer_value = \
                self.extract_fields_and_customizers(field_elements)
            field_name = field_name.strip().replace(",", "")
            if customizer_type:
                customizers[field_index] = {
                    "type": customizer_type,
                    "value": customizer_value
                }
            fields.append(self.format_type_field_name(field_name))
            field_index += 1
            column_name = alias.strip().replace(",",
                                                "") if alias else field_name
            column_names.append(self.normalize_column_name(column_name))
        return QueryElements(query_title=self.title,
                             query_text=query_text,
                             fields=fields,
                             column_names=column_names,
                             customizers=customizers,
                             resource_name=resource_name,
                             is_constant_resource=is_constant_resource)

    def cleanup_query_text(self, query: str) -> List[str]:
        query_lines = query.split("\n")
        result = []
        for line in query_lines:
            if re.match("^(#|--|//)", line):
                continue
            result.append(re.sub("(--|//).*$", "", line).strip())
        return result

    def extract_resource_from_query(self, query: str) -> str:
        return str(re.findall(r"FROM\s+(\w+)", query,
                              flags=re.IGNORECASE)[0]).strip()

    def extract_query_lines(self, query_text: str) -> List[str]:
        selected_fields = re.sub(r"\bSELECT\b|FROM .*",
                                 "",
                                 query_text,
                                 flags=re.IGNORECASE).split(",")
        return [field.strip() for field in selected_fields if field != " "]

    def extract_fields_and_aliases(
            self, query_line: str) -> Tuple[str, Optional[str]]:
        field_raw, *alias = re.split(" [Aa][Ss] ", query_line)
        return field_raw, alias[0] if alias else None

    def extract_fields_and_customizers(self, line_elements: str):
        resources = self.extract_resource_element(line_elements)
        pointers = self.extract_pointer(line_elements)
        nested_fields = self.extract_nested_resource(line_elements)
        if len(resources) > 1:
            field_name, resource_index = resources
            return field_name, "resource_index", int(resource_index)
        if len(pointers) > 1:
            field_name, pointer = pointers
            return field_name, "pointer", pointer
        if len(nested_fields) > 1:
            field_name, nested_field = nested_fields
            return field_name, "nested_field", nested_field
        return line_elements, None, None

    def extract_resource_element(self, line_elements: str) -> List[str]:
        return re.split("~", line_elements)

    def extract_pointer(self, line_elements: str) -> List[str]:
        return re.split("->", line_elements)

    def extract_nested_resource(self, line_elements: str) -> List[str]:
        return re.split(":", line_elements)

    def format_type_field_name(self, field_name: str) -> str:
        return re.sub(r"\.type", ".type_", field_name)

    def normalize_column_name(self, column_name: str) -> str:
        return re.sub(r"\.", "_", column_name)

    def normalize_query(self, query: str) -> str:
        query = self._remove_alias(query)
        query = self._remove_pointers(query)
        query = self._remove_nested_fields(query)
        query = self._clean_spaces_tabs_new_lines(query)
        query = self._remove_traling_comma(query)
        return query

    def _remove_alias(self, query: str) -> str:
        return re.sub(r"\s+[Aa][Ss]\s+(\w+)", "", query)

    def _remove_pointers(self, query: str) -> str:
        query = re.sub(r"->(\w+)|->", "", query)
        query = re.sub(r"~(\w+)|->", "", query)
        return query

    def _remove_nested_fields(self, query: str) -> str:
        return re.sub(r":((\w+)\.*){1,}", "", query)

    def _clean_spaces_tabs_new_lines(self, query: str) -> str:
        return re.sub(r"\s+", " ", query).strip()

    def _remove_traling_comma(self, query: str) -> str:
        return re.sub(r",\s+from", " FROM", query, re.IGNORECASE)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class BaseQuery:
    def __init__(self, **kwargs):
        raise NotImplementedError

    def __str__(self):
        if hasattr(self, "query_text"):
            return self.query_text
        raise NotImplementedError(
            "attribute self.query_text must be implemented "
            f"in class {self.__class__.__name__}"
        )
